using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace OmniSurf3D
{
  class OS3DAccess
  {
	//  Revision History:
	//  1.01  Added MeasureDateTime string
	  
    public Boolean WriteOmniSurf3DMap(String strFile, float[,] fSurfaceMapUM, 
	                                  String strIdentification,
									  String strMeasureDateTime,
                                      double dSpacingAlongXUM, double dSpacingAlongYUM,
                                      double dXOriginUM, double dYOriginUM,
                                      Bitmap bmpOverlay)
    {
      int nMajorVersion = 1;
      int nMinorVersion = 1;
      int nPointsAlongX = fSurfaceMapUM.GetLength(1);
      int nPointsAlongY = fSurfaceMapUM.GetLength(0);

      // verify that the destination folder exists
      String strFolder = Path.GetDirectoryName(strFile);
      try
      {
        if (!Directory.Exists(strFolder))
          Directory.CreateDirectory(strFolder);
      }
      catch // (System.Exception ex)
      {
        // cannot create directory
        return false;
      }

      FileStream fs;

      try  // attempt to open the destination file stream
      {
        fs = new FileStream(strFile, FileMode.Create);
      }
      catch // (System.Exception ex)
      {
        return false;
      }

      // attach a binary writer to the stream
      BinaryWriter bw;
      try
      {
        bw = new BinaryWriter(fs);
      }
      catch // (System.Exception ex)
      {
        fs.Close();
        return false;
      }

      // FileType - 10 characters
      char[] chFileType = "OmniSurf3D".ToCharArray();
      bw.Write(chFileType);

      // file version - 2 integers (major, minor)
      bw.Write(nMajorVersion);
      bw.Write(nMinorVersion);

      // identification - 80 characters
      int nIdentificationStringLength = 80;
      bw.Write(nIdentificationStringLength);

      String strNewIdent = strIdentification;
      if (strIdentification.Length > nIdentificationStringLength) // if too long
        strNewIdent = strIdentification.Substring(0, nIdentificationStringLength);
      strNewIdent = strNewIdent.PadRight(nIdentificationStringLength); // if too short
      bw.Write(strNewIdent.ToCharArray());

      // measure date time - 15 characters
      int nDateTimeStringLength = 15;
      bw.Write(nDateTimeStringLength);
	  
	  String strNewDateTime = strMeasureDateTime;
	  if (strMeasureDateTime.Length > nDateTimeStringLength) // if too long
		  strNewDateTime = strMeasureDateTime.Substring(0, nDateTimeStringLength);
      strNewDateTime = strNewDateTime.PadRight(nDateTimeStringLength); // if too short
      bw.Write(strNewDateTime.ToCharArray());
	  

      // double values
      bw.Write(nPointsAlongX);
      bw.Write(nPointsAlongY);
      bw.Write(dSpacingAlongXUM);
      bw.Write(dSpacingAlongYUM);
      bw.Write(dXOriginUM);
      bw.Write(dYOriginUM);

      // float heights.  float.MinValue = invalid
      for (int nY = 0; nY < nPointsAlongY; nY++)
      {
        for (int nX = 0; nX < nPointsAlongX; nX++)
        {
          bw.Write(fSurfaceMapUM[nY, nX]);
        }
      }

      Boolean bHasImage = (bmpOverlay != null);
      bw.Write(bHasImage);

      if (bHasImage)
      {
        using (MemoryStream ms = new MemoryStream())
        {
          bmpOverlay.Save(ms, System.Drawing.Imaging.ImageFormat.Png);
          ms.WriteTo(fs);
          ms.Close();
        }
      }

      bw.Flush();
      bw.Close();
      fs.Close();

      return true;
    }

    Boolean ReadOmniSurf3DMap(String strFilePath, 
                              out float[,] fSurfaceMapUM, 
                              out String strIdentification,
							  out String strMeasureDateTime,
                              out double dSpacingAlongXUM, 
                              out double dSpacingAlongYUM,
                              out double dXOriginUM, 
                              out double dYOriginUM,
                              out Bitmap bmpOverlay)
    {
      // initialize in case of abort
      fSurfaceMapUM = new float[0, 0];
      strIdentification = "";
	  strMeasureDateTime = "";
	  dSpacingAlongXUM = 1.0;
      dSpacingAlongYUM = 1.0;
      dXOriginUM = 0.0;
      dYOriginUM = 0.0;
      bmpOverlay = null;

      FileStream fs = null;
      BinaryReader br = null;

      // open the file stream
      try
      {
        fs = File.Open(strFilePath, FileMode.Open);
      }
      catch // (System.Exception ex)
      {
        return false;
      }

      // attach a binary reader to the stream
      try
      {
        br = new BinaryReader(fs);
      }
      catch // (System.Exception ex)
      {
        fs.Close();
        return false;        
      }

      // read the file type and validate
      String strFileType;
      strFileType = new String(br.ReadChars(10));
      if (strFileType != "OmniSurf3D")
      {
        br.Close();
        fs.Close();
        return false;
      }

      int nMajorVersion = 0;
      int nMinorVersion = 0;

      // file version - 2 integers (major, minor)
      nMajorVersion = br.ReadInt32();
      nMinorVersion = br.ReadInt32();
      if (nMajorVersion < 1)
      {
        br.Close();
        fs.Close();
        return false;
      }

      // read the identification and clean up
      int nIdentificationStringLength = br.ReadInt32();
      char[] chArrayIdentification = br.ReadChars(nIdentificationStringLength);
      strIdentification = new String(chArrayIdentification);
      strIdentification = strIdentification.Trim();
	  
	  // check for availability of date and time (introduced in v1.01)
	  Boolean bHasDate = (nMajorVersion >= 1) && (nMinorVersion >= 1);
	  if (bHasDate)
	  {
		int nMeasureDateTimeStringLength = br.ReadInt32();
		char[] chArrayDate = br.ReadChars(nMeasureDateTimeStringLength);
		strMeasureDateTime = new String(chArrayDate);
		strMeasureDateTime = strMeasureDateTime.Trim();
	  }
	  else
	  {
		strMeasureDateTime = "unknown";
	  }

      // read data heading information
      int nPointsAlongX = br.ReadInt32();
      int nPointsAlongY = br.ReadInt32();
      dSpacingAlongXUM = br.ReadDouble();
      dSpacingAlongYUM = br.ReadDouble();
      dXOriginUM = br.ReadDouble();
      dYOriginUM = br.ReadDouble();

      // read the surface heights (invalid = float.MinValue)
      fSurfaceMapUM = new float[nPointsAlongY, nPointsAlongX];
      for (int nY = 0; nY < nPointsAlongY; nY++)
      {
        for (int nX = 0; nX < nPointsAlongX; nX++)
        {
          fSurfaceMapUM[nY, nX] = br.ReadSingle();
        }
      }

      // check for availability of image data 
      Boolean bHasImage = br.ReadBoolean();
      if (bHasImage)
      {
        // load the image bytes into a memory stream
        MemoryStream memoryStream = new MemoryStream();
        byte[] buffer = new byte[32 * 1024]; // 32K buffer for example
        int bytesRead;
        while ((bytesRead = br.Read(buffer, 0, buffer.Length)) > 0)
        {
          memoryStream.Write(buffer, 0, bytesRead);
        }

        // convert the memory stream to an image
        memoryStream.Position = 0;
        bmpOverlay = (Bitmap)Image.FromStream(memoryStream);
      }

      br.Close();
      fs.Close();

      return true;

    }  // LoadOS3DMap
  }
}
